/*
 * deagent.c
 * WiFi DataElements Agent
 *
 * Copyright (C) 2020 IOPSYS Software Solutions AB. All rights reserved.
 *
 * Author: anjan.chanda@iopsys.eu
 *
 * See LICENSE file for license related information.
 *
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <libubus.h>
#include <libubox/utils.h>

#include <easy/easy.h>
#include <wifi.h>

#include "timer.h"
#include <cmdu_ackq.h>
#include <map_module.h>
#include "eventlog.h"

#include "wfa_delm.h"
#include "deagent.h"
#include "debug.h"



void free_radio_events(struct deagent_private *p)
{
	int i;
	struct wifi_radio_events *rv = &p->rv;

	for (i = 0; i < rv->num_radios; i++) {
		struct radio_event_data *r = rv->rdata + i;

		if (r->num_cdata > 0)
			free(r->cdata);
	}

	if (rv->num_radios > 0)
		free(rv->rdata);
}

void deagent_clear_framelist(struct deagent_private *priv)
{
	struct wifi_assoc_frame *p = NULL, *tmp;

	list_for_each_entry_safe(p, tmp,  &priv->framelist, list) {
		list_del(&p->list);
		free(p->frame);
		free(p);
	}
}

void refresh_timer_cb(atimer_t *t)
{
	struct deagent_private *p = container_of(t, struct deagent_private, t);

	fprintf(stderr, "deagent: %s: periodic = '%u' ms.\n",
		__func__, p->cfg.refresh_int);

	if (!p->getter_running)
		deagent_getter_run(p);

	if (p->cfg.refresh_int)
		timer_set(t, p->cfg.refresh_int);
}

int deagent_exit(void *priv)
{
	struct ubus_object *obj;
	struct ubus_object_type *obj_type;
	struct ubus_method *obj_methods;
	struct wifi_sta_events *sv = NULL;
	struct deagent_private *p = (struct deagent_private *)priv;

	if (p) {
		free_radio_events(p);
		ubus_unregister_event_handler(p->ctx, &p->sta_ev);
		ubus_unregister_event_handler(p->ctx, &p->wifi_radio_ev);
		ubus_unregister_event_handler(p->ctx, &p->ubus_obj_ev);

		map_unsubscribe(p->ctx, p->subscriber);

		deagent_free_dm(p->dm);

		obj = &p->obj;
		obj_methods = (struct ubus_method *)obj->methods;
		if (obj_methods)
			free(obj_methods);

		obj_type = obj->type;
		if (obj_type)
			free(obj_type);

		ubus_free(p->ctx);

		sv = p->sv;
		if (sv) {
			if (sv->num_assoc_events > 0)
				free(sv->assoclist);

			if (sv->num_disassoc_events > 0)
				free(sv->disassoclist);

			free(sv);
		}

		cmdu_ackq_free(&p->cmdu_ack_q);
		free(p);
	}

	return 0;
}

int deagent_init(void **priv, void *cfg)
{
	struct deagent_private *p;
	int ret;

	*priv = NULL;
	p = calloc(1, sizeof(struct deagent_private));
	if (!p)
		return -1;

	cmdu_ackq_init(&p->cmdu_ack_q);
	/* sta assoc frame list */
	INIT_LIST_HEAD(&p->framelist);

	p->dm = NULL;
	p->dmi = NULL;

	p->sv = calloc(1, sizeof(struct wifi_sta_events));
	if (!p->sv)
		goto out_err;

	if (cfg) {
		struct deagent_config *c = (struct deagent_config *)cfg;

		p->cfg.uctx = c->uctx;
		p->cfg.refresh_int = c->refresh_int;
		p->cfg.is_collector = 0;
		p->cfg.is_sap_mode = c->is_sap_mode;
		p->cfg.profile = c->profile;

		p->ctx = c->uctx;
	}


	fprintf(stderr, "%s:  priv = %p\n", __func__, p);
	ret = deagent_publish_object(p, WFADATAD_AGENT_OBJECT);
	if (ret)
		goto out_err;

	p->i1905 = deagent_lookup_object(p, IEEE1905_OBJECT);
	if (p->i1905 == OBJECT_INVALID) {
		err("Failed to get '%s' object!\n", IEEE1905_OBJECT);
		goto out_err;
	}

	p->i1905_map = deagent_lookup_object(p, MAP_PLUGIN_OBJECT);
	if (p->i1905_map == OBJECT_INVALID) {
		err("Failed to get '%s' object!\n", MAP_PLUGIN_OBJECT);
		goto out_err;
	}

	p->topology = deagent_lookup_object(p, "topology");
	if (p->topology == OBJECT_INVALID)
		err("Failed to get 'topology' object!\n");

	ret = deagent_get_ieee1905id(p);
	if (ret)
		goto out_err;

	ret = deagent_register_events(p);
	if (ret)
		goto out_err;

	if (!p->cfg.is_sap_mode) {
		/* subscribe for CMDU */
		deagent_subscribe_for_cmdus(p);

		/* send AP-autoconfig search message to get the controller id */
		ret = prepare_autoconfig_search(p, 0x00);	/* 2.4 GHZ */
		if (ret)
			fprintf(stderr, "error in sending ap-autoconfig message\n");
	}


	/* start DE-agent */
	deagent_getter_run(p);

	timer_init(&p->t, refresh_timer_cb);
	if (p->cfg.refresh_int)
		timer_set(&p->t, p->cfg.refresh_int);

	*priv = p;
	return 0;

out_err:
	cmdu_ackq_free(&p->cmdu_ack_q);
	deagent_clear_framelist(p);
	free(p);
	return -1;
}
