/*
 * deagent_cmdu_ackq.c
 * CMDU response and ack queue management
 *
 * Copyright (C) 2020 IOPSYS Software Solutions AB. All rights reserved.
 *
 * Author: anjan.chanda@iopsys.eu
 *
 * See LICENSE file for license related information.
 *
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <pthread.h>
#include <libubus.h>
#include <libubox/utils.h>

#include <easy/easy.h>
#include <wifi.h>

#include "timer.h"
#include <1905_tlvs.h>
#include <cmdu_ackq.h>
#include "eventlog.h"

#include "wfa_delm.h"
#include "deagent.h"

int deagent_update_stalog_data(void *sta_log, const char *ifname,
				uint8_t *mac)
{
	int ret;
	struct sta_eventlog *slog = (struct sta_eventlog *)sta_log;
	struct sta_eventlog_entry *ev = NULL;

	if (!slog->active_sta_cnt)
		return 0;

	pthread_mutex_lock(&slog->evlock);
	hlist_for_each_entry(ev, &slog->start, hlist) {
		if ((!memcmp(ev->mac, mac, 6)) && (!strcmp(ev->ifname, ifname))) {
			ret = wifi_get_sta_info(ev->ifname, ev->mac, &ev->sta);
			if (ret) {
				fprintf(stderr, "Failed to fetch STA info for " MACFMT "\n",
						MAC2STR(ev->mac));
				break;
			}
		}
	}
	pthread_mutex_unlock(&slog->evlock);

	return 0;
}

void refresh_sta_timer_cb(atimer_t *t)
{
	struct sta_eventlog_entry *msg = container_of(t, struct sta_eventlog_entry, t);
	struct sta_eventlog *sv = msg->sv;

	deagent_update_stalog_data(sv, msg->ifname, msg->mac);

	timer_set(t, 1 * 1000);
}

int sta_eventlog_init(void *sta_log)
{
	struct sta_eventlog *slog = (struct sta_eventlog *)sta_log;

	memset(slog, 0, sizeof(*slog));
	pthread_mutex_init(&slog->evlock, NULL);
	return 0;
}

struct sta_eventlog_entry *sta_eventlog_lookup(void *sta_log, uint8_t *mac)
{
	struct sta_eventlog *slog = (struct sta_eventlog *)sta_log;
	struct sta_eventlog_entry *msg = NULL;

	pthread_mutex_lock(&slog->evlock);
	hlist_for_each_entry(msg, &slog->start, hlist) {
		if (!memcmp(msg->mac, mac, 6)) {
			pthread_mutex_unlock(&slog->evlock);
			return msg;
		}
	}
	pthread_mutex_unlock(&slog->evlock);
	return NULL;
}

int sta_eventlog_enqueue(void *sta_log, const char *ifname, uint8_t *mac)
{
	struct sta_eventlog *slog = (struct sta_eventlog *)sta_log;
	struct sta_eventlog_entry *msg = NULL;

	msg = sta_eventlog_lookup(sta_log, mac);
	if (msg) {
		fprintf(stderr, "Duplicate STA event: macaddr = " MACFMT "\n",
				MAC2STR(mac));
		return -1;
	}

	msg = calloc(1, sizeof(*msg));
	if (!msg) {
		fprintf(stderr, "calloc failed. err = NOMEM\n");
		return -1;
	}

	pthread_mutex_lock(&slog->evlock);
	strncpy(msg->ifname, ifname, 15);
	memcpy(msg->mac, mac, 6);
	timer_init(&msg->t, refresh_sta_timer_cb);
	timer_set(&msg->t, 1 * 1000);
	msg->sv = slog;
	hlist_add_head(&msg->hlist, &slog->start);
	slog->active_sta_cnt++;
	fprintf(stderr, "STA ENQ: macaddr = " MACFMT "\n", MAC2STR(mac));
	pthread_mutex_unlock(&slog->evlock);

	return 0;
}

int sta_eventlog_dequeue(void *sta_log, uint8_t *mac)
{
	struct sta_eventlog *slog = (struct sta_eventlog *)sta_log;
	struct sta_eventlog_entry *msg = NULL;

	msg = sta_eventlog_lookup(sta_log, mac);
	if (!msg) {
		fprintf(stderr, "DROP! STA not found: macaddr = " MACFMT "\n",
				MAC2STR(mac));
		return -1;
	}

	pthread_mutex_lock(&slog->evlock);
	timer_del(&msg->t);
	hlist_del(&msg->hlist, &slog->start);
	slog->active_sta_cnt--;
	fprintf(stderr, "STA DEQ: macaddr = " MACFMT "\n", MAC2STR(mac));
	free(msg);
	msg = NULL;
	pthread_mutex_unlock(&slog->evlock);

	return 0;
}
