/*
 * deagent_tlv.c
 * File contains TLVs handling needed for WiFi Data Elements.
 *
 * Copyright (C) 2020 IOPSYS Software Solutions AB. All rights reserved.
 *
 * Author: saurabh.verma@iopsys.eu
 *	   anjan.chanda@iopsys.eu
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>

#ifndef _GNU_SOURCE
#define _GNU_SOURCE
#endif

#include <json-c/json.h>
#include <libubox/blobmsg.h>
#include <libubox/blobmsg_json.h>
#include <libubox/uloop.h>
#include <libubox/ustream.h>
#include <libubox/utils.h>
#include <libubus.h>

#include <easy/easy.h>
#include <wifi.h>

#include "timer.h"
#include <cmdu.h>
#include <1905_tlvs.h>
#include <map2.h>
#include <cmdu_ackq.h>

#include "eventlog.h"
#include "wfa_delm.h"
#include "deagent.h"

#include "bufutil.h"	//TODO: remove

int deagent_gen_aladdr(struct deagent_private *priv, struct cmdu_buff *cmdu)
{
	int ret;
	struct tlv *t;
	struct tlv_aladdr *data;

	t = cmdu_reserve_tlv(cmdu, 30);
	if (!t)
		return -1;

	t->type = TLV_TYPE_AL_MAC_ADDRESS_TYPE;
	t->len = sizeof(*data);
	data = (struct tlv_aladdr *)t->data;
	memcpy(data->macaddr, priv->ieee1905_macaddr, 6);

	ret = cmdu_put_tlv(cmdu, t);
	if (ret) {
		fprintf(stderr, "%s: error: cmdu_put_tlv()\n", __func__);
		return -1;
	}

	return 0;
}

int deagent_gen_searched_role(struct deagent_private *priv,
		struct cmdu_buff *cmdu, uint8_t role)
{
	int ret;
	struct tlv *t;
	struct tlv_searched_role *data;

	t = cmdu_reserve_tlv(cmdu, 30);
	if (!t)
		return -1;

	t->type = TLV_TYPE_SEARCHED_ROLE;
	t->len = sizeof(*data);
	data = (struct tlv_searched_role *)t->data;
	data->role = role;

	ret = cmdu_put_tlv(cmdu, t);
	if (ret) {
		fprintf(stderr, "%s: error: cmdu_put_tlv()\n", __func__);
		return -1;
	}

	return 0;
}

int deagent_gen_autoconfig_band(struct deagent_private *priv,
		struct cmdu_buff *cmdu, uint8_t band)
{
	int ret;
	struct tlv *t;
	struct tlv_autoconfig_band *data;

	t = cmdu_reserve_tlv(cmdu, 30);
	if (!t)
		return -1;

	t->type = TLV_TYPE_AUTOCONFIG_FREQ_BAND;
	t->len = sizeof(*data);
	data = (struct tlv_autoconfig_band *)t->data;
	data->band = band;

	ret = cmdu_put_tlv(cmdu, t);
	if (ret) {
		fprintf(stderr, "%s: error: cmdu_put_tlv()\n", __func__);
		return -1;
	}

	return 0;
}

int deagent_gen_supported_service(struct deagent_private *priv,
		struct cmdu_buff *cmdu, uint8_t service)
{
	int i, ret;
	struct tlv *t;

	t = cmdu_reserve_tlv(cmdu, 30);
	if (!t)
		return -1;

	t->type = MAP_TLV_SUPPORTED_SERVICE;
	t->data[0] = 1;
	for (i = 0; i < t->data[0]; i++)
		t->data[1 + i] = service;

	t->len = 1 + t->data[0];

	ret = cmdu_put_tlv(cmdu, t);
	if (ret) {
		fprintf(stderr, "%s: error: cmdu_put_tlv()\n", __func__);
		return -1;
	}

	return 0;
}

int deagent_gen_searched_service(struct deagent_private *priv,
		struct cmdu_buff *cmdu, uint8_t service)
{
	int i, ret;
	struct tlv *t;

	t = cmdu_reserve_tlv(cmdu, 30);
	if (!t)
		return -1;

	t->type = MAP_TLV_SEARCHED_SERVICE;
	t->data[0] = 1;
	for (i = 0; i < t->data[0]; i++)
		t->data[1 + i] = service;

	t->len = 1 + t->data[0];

	ret = cmdu_put_tlv(cmdu, t);
	if (ret) {
		fprintf(stderr, "%s: error: cmdu_put_tlv()\n", __func__);
		return -1;
	}

	return 0;
}

int deagent_gen_map_profile(struct deagent_private *priv,
		struct cmdu_buff *cmdu, uint8_t profile)
{
	int ret;
	struct tlv *t;
	struct tlv_map_profile *data;

	t = cmdu_reserve_tlv(cmdu, 20);
	if (!t)
		return -1;

	t->type = MAP_TLV_MULTIAP_PROFILE;
	t->len = sizeof(*data);
	data = (struct tlv_map_profile *)t->data;
	data->profile = profile;

	ret = cmdu_put_tlv(cmdu, t);
	if (ret) {
		fprintf(stderr, "%s: error: cmdu_put_tlv()\n", __func__);
		return -1;
	}

	return 0;
}

int deagent_gen_i1905_tlvs(struct wifi_network_device *dev,
		struct cmdu_buff *cmdu)
{
	int ret;
	uint8_t *tlv = NULL;
	uint8_t *cur_tlv = NULL;
	int len = 0;
	int datalen;
	int copy_len;

	if (!dev->ieee1905_tlv_data)
		return -1;

	cur_tlv = dev->ieee1905_tlv_data;
	datalen = dev->ieee1905_tlv_datalen;
	while (len < datalen) {
		tlv = cur_tlv + len;
		copy_len = tlv_total_length((struct tlv *)tlv);

		switch (*tlv) {
		case TLV_TYPE_DEVICE_INFORMATION_TYPE:
		case TLV_TYPE_DEVICE_BRIDGING_CAPABILITIES:
		case TLV_TYPE_NON_1905_NEIGHBOR_DEVICE_LIST:
		case TLV_TYPE_NEIGHBOR_DEVICE_LIST:

			ret = cmdu_put(cmdu, tlv, copy_len);
			if (ret) {
				fprintf(stderr, "%s: error: cmdu_put()\n",
						__func__);
				return -1;
			}
		}
		len += copy_len;
	}

	return 0;
}

int deagent_gen_ap_oper_bss(struct wifi_network_device *dev,
		struct cmdu_buff *cmdu)
{
	int i, j;
	int ret;
	int offset = 0;
	struct wifi_radio_element *radio;
	struct wifi_bss_element *bss;
	struct tlv *t;
	struct tlv_ap_oper_bss *data;
	int reserve_len = 512;

	t = cmdu_reserve_tlv(cmdu, reserve_len);
	if (!t)
		return -1;

	t->type = MAP_TLV_AP_OPERATIONAL_BSS;
	data = (struct tlv_ap_oper_bss *)t->data;
	data->num_radio = dev->num_radios;
	offset = sizeof(*data);
	for (i = 0; i < data->num_radio; i++) {
		struct ap_oper_bss_radiolist *r =
			(struct ap_oper_bss_radiolist *)&t->data[offset];

		if ((offset + 6 + 1) > reserve_len)
			goto error;

		radio = dev->radiolist + i;
		memcpy(r->radio, radio->macaddr, 6);
		offset += sizeof(*r);

		for (j = 0; j < radio->num_bss; j++) {
			int len;
			struct ap_oper_bss_bss *b =
				(struct ap_oper_bss_bss *)&t->data[offset];

			if ((offset + 6 + 1) > reserve_len)
				goto error;

			bss = radio->bsslist + j;
			if (hwaddr_is_zero(bss->bssid))
				continue;
			r->num_bss++;
			memcpy(b->bssid, bss->bssid, 6);
			len = strlen(bss->ssid);
			if ((offset + 7 + len) > reserve_len)
				goto error;

			memcpy(b->ssid, bss->ssid, len);
			b->ssidlen = len;
			offset += sizeof(*b) + len;
		}
	}

	/* update the tlv length */
	t->len = offset;

	ret = cmdu_put_tlv(cmdu, t);
	if (ret) {
		fprintf(stderr, "%s: error: cmdu_put_tlv()\n", __func__);
		return -1;
	}

	return 0;

error:
	return -1;
}

int deagent_gen_assoc_client(struct wifi_network_device *dev,
		struct cmdu_buff *cmdu)
{
	int i, j, k;
	int ret;
	int offset = 0;
	struct tlv *t;
	struct tlv_assoc_client *data;
	struct wifi_radio_element *radio;
	struct wifi_bss_element *bss;
	struct wifi_sta_element *sta;
	int reserve_len = 512;

	t = cmdu_reserve_tlv(cmdu, reserve_len);
	if (!t)
		return -1;

	t->type = MAP_TLV_ASSOCIATED_CLIENTS;
	data = (struct tlv_assoc_client *)t->data;
	data->num_bss = 0;
	offset = sizeof(*data);
	for (i = 0; i < dev->num_radios; i++) {
		radio = dev->radiolist + i;

		//data->num_bss += radio->num_bss;
		for (j = 0; j < radio->num_bss; j++) {
			struct assoc_client_bss *b =
				(struct assoc_client_bss *)&t->data[offset];

			bss = radio->bsslist + j;
			if ((offset + 6 + 2) > reserve_len)
				goto error;

			if (hwaddr_is_zero(bss->bssid))
				continue;

			data->num_bss++;
			memcpy(b->bssid, bss->bssid, 6);
			BUF_PUT_BE16(b->num_client,
					bss->num_stations & 0xffff);
			offset += sizeof(*b);
			for (k = 0; k < bss->num_stations; k++) {
				struct assoc_client_sta *s =
					(struct assoc_client_sta *)&t->data[offset];

				sta = bss->stalist + k;
				if ((offset + 6 + 2) > reserve_len)
					goto error;

				memcpy(s->macaddr, sta->macaddr, 6);
				BUF_PUT_BE16(s->conntime, sta->conn_time);
				offset += sizeof(*s);
			}
		}
	}

	/* update the TLV length */
	t->len = offset;

	ret = cmdu_put_tlv(cmdu, t);
	if (ret) {
		fprintf(stderr, "%s: error: cmdu_put_tlv()\n", __func__);
		return -1;
	}

	return 0;

error:
	return -1;
}

int deagent_gen_ap_cap(struct wifi_network_device *dev, struct cmdu_buff *cmdu)
{
	int ret;
	struct tlv *t;
	struct tlv_ap_cap *data;

	t = cmdu_reserve_tlv(cmdu, 20);
	if (!t)
		return -1;

	t->type = MAP_TLV_AP_CAPABILITY;
	t->len = sizeof(*data);
	data = (struct tlv_ap_cap *)t->data;
	/* Support Agent-initiated RCPI-based Steering */
	data->cap |= AGENT_SUPPORTS_RCPI_STEER;

	ret = cmdu_put_tlv(cmdu, t);
	if (ret) {
		fprintf(stderr, "%s: error: cmdu_put_tlv()\n", __func__);
		return -1;
	}

	return 0;
}

int deagent_gen_ap_radio_basic_cap(struct wifi_network_device *dev,
		struct cmdu_buff *cmdu, int radio_index)
{
	int i, ret;
	int offset = 0;
	struct tlv *t;
	struct tlv_ap_radio_basic_cap *data;
	struct wifi_radio_element *radio;
	int reserve_len = 512;

	t = cmdu_reserve_tlv(cmdu, reserve_len);
	if (!t)
		return -1;

	t->type = MAP_TLV_AP_RADIO_BASIC_CAPABILITIES;
	data = (struct tlv_ap_radio_basic_cap *)t->data;

	radio = dev->radiolist + radio_index;
	memcpy(data->radio, radio->macaddr, 6);
	data->max_bssnum = 4; /*TODO: max bss per radio */
	data->num_opclass = radio->num_supp_opclass;
	offset = sizeof(*data);
	for (i = 0; i < data->num_opclass; i++) {
		struct ap_radio_basic_cap_opclass *op =
			(struct ap_radio_basic_cap_opclass *)&t->data[offset];

		if ((offset + 3) > reserve_len)
			goto error;

		op->classid = radio->supp_opclass[i].id;
		op->max_txpower = radio->supp_opclass[i].max_txpower;
		op->num_nonop_channel =
			radio->supp_opclass[i].num_exclude_channels;
		if (op->num_nonop_channel > 0) {
			if ((offset + 3 + op->num_nonop_channel) > reserve_len)
				goto error;

			memcpy(op->nonop_channel,
					radio->supp_opclass[i].exclude_chanlist,
					op->num_nonop_channel);
		}

		offset += sizeof(*op) + op->num_nonop_channel;
	}

	/* update the tlv length */
	t->len = offset;

	ret = cmdu_put_tlv(cmdu, t);
	if (ret) {
		fprintf(stderr, "%s: error: cmdu_put_tlv()\n", __func__);
		return -1;
	}

	return 0;

error:
	return -1;
}

int deagent_gen_ap_ht_cap(struct wifi_network_device *dev,
		struct cmdu_buff *cmdu, int radio_index)
{
	int ret;
	struct tlv *t;
	struct tlv_ap_ht_cap *data;
	struct wifi_radio_element *radio;

	t = cmdu_reserve_tlv(cmdu, 20);
	if (!t)
		return -1;

	t->type = MAP_TLV_AP_HT_CAPABILITIES;
	t->len = sizeof(*data);
	data = (struct tlv_ap_ht_cap *)t->data;
	radio = dev->radiolist + radio_index;
	memcpy(data->radio, radio->macaddr, 6);
	memcpy(&data->cap, &radio->caps.ht, 1);

	ret = cmdu_put_tlv(cmdu, t);
	if (ret) {
		fprintf(stderr, "%s: error: cmdu_put_tlv()\n", __func__);
		return -1;
	}

	return 0;
}

int deagent_gen_ap_vht_cap(struct wifi_network_device *dev,
		struct cmdu_buff *cmdu, int radio_index)
{
	int ret;
	struct tlv *t;
	struct tlv_ap_vht_cap *data;
	struct wifi_radio_element *radio;

	t = cmdu_reserve_tlv(cmdu, 30);
	if (!t)
		return -1;

	t->type = MAP_TLV_AP_VHT_CAPABILITIES;
	t->len = sizeof(*data);
	data = (struct tlv_ap_vht_cap *)t->data;
	radio = dev->radiolist + radio_index;
	memcpy(data->radio, radio->macaddr, 6);
	data->tx_mcs_supported = BUF_GET_BE16(radio->caps.vht[0]);
	data->rx_mcs_supported = BUF_GET_BE16(radio->caps.vht[2]);
	memcpy(data->cap, &radio->caps.vht[4], 2);

	ret = cmdu_put_tlv(cmdu, t);
	if (ret) {
		fprintf(stderr, "%s: error: cmdu_put_tlv()\n", __func__);
		return -1;
	}

	return 0;
}

int deagent_gen_ap_he_cap(struct wifi_network_device *dev,
		struct cmdu_buff *cmdu, int radio_index)
{
	int ret;
	int offset = 0;
	int index, mcs_len;
	struct tlv *t;
	struct wifi_radio_element *radio;
	int reserve_len = 64;

	t = cmdu_reserve_tlv(cmdu, reserve_len);
	if (!t)
		return -1;

	t->type = MAP_TLV_AP_HE_CAPABILITIES;

	radio = dev->radiolist + radio_index;
	memcpy(&t->data[offset], radio->macaddr, 6);
	offset += 6;
	t->data[offset++] = radio->caps.he[0];
	mcs_len = radio->caps.he[0];

	if ((offset + mcs_len + 2) > reserve_len)
		return -1;

	if (mcs_len > 0)
		memcpy(&t->data[offset], &radio->caps.he[1], mcs_len);

	offset += mcs_len;
	index = 1 + mcs_len;
	memcpy(&t->data[offset], &radio->caps.he[index], 2);

	t->len = offset + 2;

	ret = cmdu_put_tlv(cmdu, t);
	if (ret) {
		fprintf(stderr, "%s: error: cmdu_put_tlv()\n", __func__);
		return -1;
	}

	return 0;
}

int deagent_gen_channel_scan_capability(struct wifi_network_device *dev,
		struct cmdu_buff *cmdu)
{
	int ret;
	int i, j;
	int offset = 0;
	struct tlv *t;
	struct tlv_channel_scan_capability *data;
	struct wifi_radio_element *radio;
	struct wifi_opclass_supported_element *sop;
	int reserve_len = 512;

	t = cmdu_reserve_tlv(cmdu, reserve_len);
	if (!t)
		return -1;

	t->type = MAP_TLV_CHANNEL_SCAN_CAPABILITY;
	data = (struct tlv_channel_scan_capability *)t->data;
	data->num_radio = dev->num_radios;
	offset = sizeof(*data);

	for (i = 0; i < dev->num_radios; i++) {
		struct channel_scan_capability_radio *r =
			(struct channel_scan_capability_radio *)&t->data[offset];

		if ((offset + sizeof(*r)) > reserve_len)
			goto error;

		radio = dev->radiolist + i;
		memcpy(r->radio, radio->macaddr, 6);
		/* Perform requested scans */
		r->cap |= SCAN_CAP_IMPACT;
		BUF_PUT_BE32(r->min_scan_interval, 30);
		r->num_opclass = radio->num_supp_opclass;
		offset += sizeof(*r);

		for (j = 0; j < radio->num_supp_opclass; j++) {
			struct channel_scan_capability_opclass *op =
				(struct channel_scan_capability_opclass *)&t->data[offset];

			if ((offset + sizeof(*op)) > reserve_len)
				goto error;

			sop = radio->supp_opclass + j;
			op->classid = sop->id;
			op->num_channel = 0; /* Scan all channels */
			offset += sizeof(*op);

			if ((offset + op->num_channel) > reserve_len)
				goto error;

			if (op->num_channel > 0) {
				offset += op->num_channel;
				memcpy(op->channel, sop->chanlist,
						op->num_channel);
			}
		}
	}

	/* update the tlv length */
	t->len = offset;

	ret = cmdu_put_tlv(cmdu, t);
	if (ret) {
		fprintf(stderr, "%s: error: cmdu_put_tlv()\n", __func__);
		return -1;
	}

	return 0;

error:
	return -1;
}

static inline void put_duration_data(uint8_t *buf, uint32_t val)
{
	/* ignore val MSB data */
	buf[0] = (val >> 16) & 0x0ff;
	buf[1] = (val >> 8) & 0xff;
	buf[2] = val & 0xff;
}

int deagent_gen_cac_cap(struct wifi_network_device *dev,
		struct cmdu_buff *cmdu)
{
	int i, j, k, l;
	int ret;
	int offset = 0;
	uint8_t ch;
	struct tlv *t;
	struct tlv_cac_cap *data;
	struct wifi_radio_element *radio;
	struct wifi_opclass_supported_element *sop;
	int reserve_len = 512;

	t = cmdu_reserve_tlv(cmdu, reserve_len);
	if (!t)
		return -1;

	t->type = MAP_TLV_CAC_CAPABILITY;
	data = (struct tlv_cac_cap *)t->data;

	memcpy(data->country, dev->radiolist[0].country_code, 2);
	data->num_radio = dev->num_radios;
	offset = sizeof(*data);
	for (i = 0; i < data->num_radio; i++) {
		struct cac_cap_radio *r =
			(struct cac_cap_radio *)&t->data[offset];

		if ((offset + 7) > reserve_len)
			goto error;

		radio = dev->radiolist + i;
		memcpy(r->radio, radio->macaddr, 6);
		r->num_cac = 1;
		offset += sizeof(*r);
		for (j = 0; j < r->num_cac; j++) {
			struct cac_cap_cac *c =
				(struct cac_cap_cac *)&t->data[offset];

			if ((offset + 5) > reserve_len)
				goto error;

			c->supp_method |= CAC_METHOD_CONTINUOUS_CAC;
			/* Default duration is 60,,, 600 for channels (120, 124 & 128).
			 * TODO: Assumption, All the supporting op_class supported continuous CAC method,
			 * Currently no parmaeter/method to classify the same
			 */
			/* TODO/revisit for putting data in duration buffer, 3 byte */
			put_duration_data(c->duration, 60);
			c->num_opclass = radio->num_supp_opclass;
			offset += sizeof(*c);
			for (k = 0; k < radio->num_supp_opclass; k++) {
				struct cac_cap_opclass *op =
					(struct cac_cap_opclass *)&t->data[offset];

				sop = radio->supp_opclass + k;
				if ((offset + 2 + sop->num_channels) > reserve_len)
					goto error;

				op->classid = sop->id;
				op->num_channel = sop->num_channels;
				offset += sizeof(*op) + op->num_channel;

				for (l = 0; l < sop->num_channels; l++) {
					op->channel[l] = sop->chanlist[l];
					ch = op->channel[l];
					if (ch == 120 || ch == 124 || ch == 128)
						put_duration_data(c->duration, 600);
				}
			}
		}
	}

	/* update the tlv length */
	t->len = offset;

	ret = cmdu_put_tlv(cmdu, t);
	if (ret) {
		fprintf(stderr, "%s: error: cmdu_put_tlv()\n", __func__);
		return -1;
	}

	return 0;

error:
	return -1;
}

int deagent_gen_profile2_ap_cap(struct wifi_network_device *dev,
		struct cmdu_buff *cmdu)
{
	int ret;
	struct tlv *t;
	struct tlv_profile2_ap_cap *data;

	t = cmdu_reserve_tlv(cmdu, 20);
	if (!t)
		return -1;

	t->type = MAP_TLV_PROFILE2_AP_CAPABILITY;
	t->len = sizeof(*data);
	data = (struct tlv_profile2_ap_cap *)t->data;
	data->unit |= STATS_UNIT_BYTE;
	data->max_vids = (uint8_t)255;

	ret = cmdu_put_tlv(cmdu, t);
	if (ret) {
		fprintf(stderr, "%s: error: cmdu_put_tlv()\n", __func__);
		return -1;
	}

	return 0;
}

#define AP_COLLECTION_INTERVAL	(10 * 1000)
int deagent_gen_metric_collection_int(struct wifi_network_device *dev,
		struct cmdu_buff *cmdu)
{
	int ret;
	struct tlv *t;
	struct tlv_metric_collection_int *data;

	t = cmdu_reserve_tlv(cmdu, 20);
	if (!t)
		return -1;

	t->type = MAP_TLV_METRIC_COLLECTION_INTERVAL;
	t->len = sizeof(*data);
	data = (struct tlv_metric_collection_int *)t->data;
	BUF_PUT_BE32(data->interval, AP_COLLECTION_INTERVAL);

	ret = cmdu_put_tlv(cmdu, t);
	if (ret) {
		fprintf(stderr, "%s: error: cmdu_put_tlv()\n", __func__);
		return -1;
	}

	return 0;
}

int deagent_gen_radio_metrics(struct wifi_network_device *dev,
		struct cmdu_buff *cmdu, int radio_index)
{
	int ret;
	struct tlv *t;
	struct tlv_radio_metrics *data;
	struct wifi_radio_element *radio;

	t = cmdu_reserve_tlv(cmdu, 30);
	if (!t)
		return -1;

	t->type = MAP_TLV_RADIO_METRICS;
	t->len = sizeof(*data);
	data = (struct tlv_radio_metrics *)t->data;
	radio = dev->radiolist + radio_index;

	memcpy(data->radio, radio->macaddr, 6);
	data->noise = radio->anpi;
	data->transmit = radio->tx_utilization;
	data->receive_self = radio->rx_utilization;
	data->receive_other = radio->other_utilization;

	ret = cmdu_put_tlv(cmdu, t);
	if (ret) {
		fprintf(stderr, "%s: error: cmdu_put_tlv()\n", __func__);
		return -1;
	}

	return 0;
}

int deagent_gen_ap_metrics(struct wifi_network_device *dev,
		struct cmdu_buff *cmdu, int radio_index, int bss_index)
{
	int ret;
	int copy_index = 0;
	struct tlv *t;
	struct tlv_ap_metrics *data;
	struct wifi_radio_element *radio = dev->radiolist + radio_index;
	struct wifi_bss_element *bss = radio->bsslist + bss_index;

	t = cmdu_reserve_tlv(cmdu, 64);
	if (!t)
		return -1;

	t->type = MAP_TLV_AP_METRICS;
	t->len = sizeof(*data);
	data = (struct tlv_ap_metrics *)t->data;

	memcpy(data->bssid, bss->bssid, 6);
	// TODO:FIXME, channel_utilization
	data->channel_utilization = 0;
	BUF_PUT_BE16(data->num_station, bss->num_stations);
	data->esp_ac = ESP_AC_BE;
	memcpy(data->esp_be, bss->est_wmm_be, 3);

	if (bss->is_ac_bk) {
		data->esp_ac |= ESP_AC_BK;
		memcpy(data->esp + copy_index, bss->est_wmm_bk, 3);
		copy_index += 3;
	}

	if (bss->is_ac_vo) {
		data->esp_ac |= ESP_AC_VO;
		memcpy(data->esp + copy_index, bss->est_wmm_vo, 3);
		copy_index += 3;
	}

	if (bss->is_ac_vi) {
		data->esp_ac |= ESP_AC_VI;
		memcpy(data->esp + copy_index, bss->est_wmm_vi, 3);
		copy_index += 3;
	}
	t->len += copy_index;

	ret = cmdu_put_tlv(cmdu, t);
	if (ret) {
		fprintf(stderr, "%s: error: cmdu_put_tlv()\n", __func__);
		return -1;
	}

	return 0;
}

int deagent_gen_ap_ext_metrics(struct wifi_network_device *dev,
		struct cmdu_buff *cmdu, int radio_index, int bss_index)
{
	int ret;
	struct tlv *t;
	struct tlv_ap_ext_metrics *data;
	struct wifi_radio_element *radio = dev->radiolist + radio_index;
	struct wifi_bss_element *bss = radio->bsslist + bss_index;

	t = cmdu_reserve_tlv(cmdu, 64);
	if (!t)
		return -1;

	t->type = MAP_TLV_AP_EXTENDED_METRICS;
	t->len = sizeof(*data);
	data = (struct tlv_ap_ext_metrics *)t->data;

	memcpy(data->bssid, bss->bssid, 6);
	BUF_PUT_BE32(data->tx_bytes_ucast, bss->tx_ucast_bytes);
	BUF_PUT_BE32(data->rx_bytes_ucast, bss->rx_ucast_bytes);
	BUF_PUT_BE32(data->tx_bytes_mcast, bss->tx_mcast_bytes);
	BUF_PUT_BE32(data->rx_bytes_mcast, bss->rx_mcast_bytes);
	BUF_PUT_BE32(data->tx_bytes_bcast, bss->tx_bcast_bytes);
	BUF_PUT_BE32(data->rx_bytes_bcast, bss->rx_bcast_bytes);

	ret = cmdu_put_tlv(cmdu, t);
	if (ret) {
		fprintf(stderr, "%s: error: cmdu_put_tlv()\n", __func__);
		return -1;
	}

	return 0;
}

int deagent_gen_assoc_sta_traffic_stats(struct wifi_network_device *dev,
		struct cmdu_buff *cmdu, int radio_index,
		int bss_index, int sta_index)
{
	int ret;
	struct tlv *t;
	struct tlv_assoc_sta_traffic_stats *data;
	struct wifi_radio_element *radio = dev->radiolist + radio_index;
	struct wifi_bss_element *bss = radio->bsslist + bss_index;
	struct wifi_sta_element *sta = bss->stalist + sta_index;

	t = cmdu_reserve_tlv(cmdu, 64);
	if (!t)
		return -1;

	t->type = MAP_TLV_ASSOCIATED_STA_TRAFFIC_STATS;
	t->len = sizeof(*data);
	data = (struct tlv_assoc_sta_traffic_stats *)t->data;

	memcpy(data->macaddr, sta->macaddr, 6);
	BUF_PUT_BE32(data->tx_bytes, sta->tx_bytes);
	BUF_PUT_BE32(data->rx_bytes, sta->rx_bytes);
	BUF_PUT_BE32(data->tx_packets, sta->tx_pkts);
	BUF_PUT_BE32(data->rx_packets, sta->rx_pkts);
	BUF_PUT_BE32(data->tx_err_packets, sta->tx_errors);
	BUF_PUT_BE32(data->tx_err_packets, sta->rx_errors);
	BUF_PUT_BE32(data->rtx_packets, sta->rtx_pkts);

	ret = cmdu_put_tlv(cmdu, t);
	if (ret) {
		fprintf(stderr, "%s: error: cmdu_put_tlv()\n", __func__);
		return -1;
	}

	return 0;
}

int deagent_gen_assoc_sta_link_metrics(struct wifi_network_device *dev,
		struct cmdu_buff *cmdu, int radio_index,
		int bss_index, int sta_index)
{
	int ret, i;
	int offset = 0;
	struct tlv *t;
	struct tlv_assoc_sta_link_metrics *data;
	struct wifi_radio_element *radio = dev->radiolist + radio_index;
	struct wifi_bss_element *bss = radio->bsslist + bss_index;
	struct wifi_sta_element *sta = bss->stalist + sta_index;
	int reserve_len = 256;

	t = cmdu_reserve_tlv(cmdu, reserve_len);
	if (!t)
		return -1;

	t->type = MAP_TLV_ASSOCIATED_STA_LINK_METRICS;
	data = (struct tlv_assoc_sta_link_metrics *)t->data;
	memcpy(data->macaddr, sta->macaddr, 6);
	data->num_bss = 1;
	offset = sizeof(*data);

	for (i = 0; i < data->num_bss; i++) {
		struct assoc_sta_link_metrics_bss *b =
			(struct assoc_sta_link_metrics_bss *)&t->data[offset];

		if ((offset + 19) > reserve_len)
			return -1;

		memcpy(b->bssid, bss->bssid, 6);
		BUF_PUT_BE32(b->time_delta, sta->conn_time);
		BUF_PUT_BE32(b->dl_thput, sta->dl_est_thput);
		BUF_PUT_BE32(b->ul_thput, sta->ul_est_thput);
		b->ul_rcpi = sta->rssi;
		offset += sizeof(*b);
	}

	/* update the tlv length */
	t->len = offset;

	ret = cmdu_put_tlv(cmdu, t);
	if (ret) {
		fprintf(stderr, "%s: error: cmdu_put_tlv()\n", __func__);
		return -1;
	}

	return 0;
}

int deagent_gen_assoc_sta_ext_link_metric(struct wifi_network_device *dev,
		struct cmdu_buff *cmdu, int radio_index,
		int bss_index, int sta_index)
{
	int ret, i;
	int offset = 0;
	struct tlv *t;
	struct tlv_sta_ext_link_metric *data;
	struct wifi_radio_element *radio = dev->radiolist + radio_index;
	struct wifi_bss_element *bss = radio->bsslist + bss_index;
	struct wifi_sta_element *sta = bss->stalist + sta_index;
	int reserve_len = 256;

	t = cmdu_reserve_tlv(cmdu, reserve_len);
	if (!t)
		return -1;

	t->type = MAP_TLV_ASSOCIATED_STA_EXT_LINK_METRICS;
	data = (struct tlv_sta_ext_link_metric *)t->data;
	memcpy(data->macaddr, sta->macaddr, 6);
	data->num_bss = 1;
	offset = sizeof(*data);

	for (i = 0; i < data->num_bss; i++) {
		struct sta_ext_link_metric_bss *b =
			(struct sta_ext_link_metric_bss *)&t->data[offset];

		if ((offset + 22) > reserve_len)
			return -1;

		memcpy(b->bssid, bss->bssid, 6);
		BUF_PUT_BE32(b->dl_rate, sta->dl_rate);
		BUF_PUT_BE32(b->ul_rate, sta->ul_rate);
		BUF_PUT_BE32(b->rx_util, sta->dl_utilization);
		BUF_PUT_BE32(b->tx_util, sta->ul_utilization);
		offset += sizeof(*b);
	}

	/* update the tlv length */
	t->len = offset;

	ret = cmdu_put_tlv(cmdu, t);
	if (ret) {
		fprintf(stderr, "%s: error: cmdu_put_tlv()\n", __func__);
		return -1;
	}

	return 0;
}

int deagent_gen_bsta_radio_cap(struct wifi_network_device *dev,
		struct cmdu_buff *cmdu, int radio_index)
{
	int ret;
	struct tlv *t;
	struct tlv_bsta_radio_cap *data;
	struct wifi_radio_element *radio = dev->radiolist + radio_index;

	t = cmdu_reserve_tlv(cmdu, 256);
	if (!t)
		return -1;

	t->type = MAP_TLV_BACKHAUL_STA_RADIO_CAPABILITY;
	t->len = sizeof(*data);
	data = (struct tlv_bsta_radio_cap *)t->data;

	memcpy(data->radio, radio->macaddr, 6);

	if (!hwaddr_is_zero(radio->bksta.macaddr)) {
		data->macaddr_included = BSTA_MACADDRESS_INCLUDED;
		memcpy(&data->macaddr[0], radio->bksta.macaddr, 6);
		t->len += sizeof(macaddr_t);
	}

	ret = cmdu_put_tlv(cmdu, t);
	if (ret) {
		fprintf(stderr, "%s: error: cmdu_put_tlv()\n", __func__);
		return -1;
	}

	return 0;
}

int deagent_gen_oper_channel_report(struct wifi_network_device *dev,
		struct cmdu_buff *cmdu, int radio_index)
{
	int i, ret;
	int offset = 0;
	uint8_t txpower = 0;
	struct tlv *t;
	struct wifi_radio_element *radio;
	int reserve_len = 256;

	t = cmdu_reserve_tlv(cmdu, reserve_len);
	if (!t)
		return -1;

	radio = dev->radiolist + radio_index;
	t->type = MAP_TLV_OPERATING_CHANNEL_REPORT;
	/* radio id */
	memcpy(&t->data[offset], radio->macaddr, 6);
	offset += 6;
	/* num opclass */
	t->data[offset++] = radio->num_curr_opclass;

	for (i = 0; i < radio->num_curr_opclass; i++) {
		if ((offset + 3) > reserve_len)
			return -1;

		/* class id */
		t->data[offset++] = radio->curr_opclass[i].id;
		/* channel */
		t->data[offset++] = radio->curr_opclass[i].channel;
		txpower = radio->curr_opclass[i].txpower;
	}
	/* txpower */
	t->data[offset++] = txpower;

	/* update the tlv length */
	t->len = offset;

	ret = cmdu_put_tlv(cmdu, t);
	if (ret) {
		fprintf(stderr, "%s: error: cmdu_put_tlv()\n", __func__);
		return -1;
	}

	return 0;
}

int deagent_gen_timestamp(struct wifi_network_device *dev,
		struct cmdu_buff *cmdu)
{
	int ret;
	struct tlv *t;
	char tsp[32] = {0};
	struct tlv_timestamp *data;

	t = cmdu_reserve_tlv(cmdu, 64);
	if (!t)
		return -1;

	t->type = MAP_TLV_TIMESTAMP;
	data = (struct tlv_timestamp *)t->data;
	get_timestamp(NULL, tsp);
	data->len = strlen(tsp);
	memcpy(data->timestamp, (uint8_t *)tsp, data->len);
	t->len = sizeof(*data) + data->len;

	ret = cmdu_put_tlv(cmdu, t);
	if (ret) {
		fprintf(stderr, "%s: error: cmdu_put_tlv()\n", __func__);
		return -1;
	}

	return 0;
}

int deagent_gen_channel_scan_result(struct wifi_network_device *dev,
		struct cmdu_buff *cmdu, int radio_index,
		int opclass_index, int channel_index)
{
	int i, ret;
	int offset = 0;
	uint8_t status = CH_SCAN_STATUS_SUCCESS;
	uint8_t bssload_elem_pres = CH_SCAN_RESULT_BSSLOAD_PRESENT;
	char bw_str[16] = {0};
	struct tlv *t;
	struct wifi_radio_element *radio = dev->radiolist + radio_index;
	struct wifi_scanres_opclass_element *op =
		radio->scanlist[0].opclass_scanlist + opclass_index;
	struct wifi_scanres_channel_element *ch =
		op->channel_scanlist + channel_index;
	int reserve_len = 200;

	/* send the valid channel data */
	if  (ch->channel <= 0)
		return 0;

	t = cmdu_reserve_tlv(cmdu, reserve_len);
	if (!t)
		return -1;

	t->type = MAP_TLV_CHANNEL_SCAN_RES;
	/* radio id */
	memcpy(&t->data[offset], radio->macaddr, 6);
	offset += 6;
	t->data[offset++] = op->opclass;	/* opclass */
	t->data[offset++] = ch->channel;	/* channel */
	t->data[offset++] = status;		/*status */

	if (status != CH_SCAN_STATUS_SUCCESS)
		goto out;

	t->data[offset++] = strlen(ch->tsp);			/* timestamp len */
	if ((offset + strlen(ch->tsp) + 4) > reserve_len)
		goto error;

	memcpy(&t->data[offset], ch->tsp, strlen(ch->tsp));	/* timestamp */
	offset += strlen(ch->tsp);
	t->data[offset++] = ch->utilization;	/* utilization */
	t->data[offset++] = ch->anpi;		/* noise */

	/* num neighbors */
	BUF_PUT_BE16(t->data[offset], ch->num_neighbors);
	offset += 2;
	for (i = 0; i < ch->num_neighbors; i++) {
		struct wifi_scanres_neighbor_element *nbr = ch->nbrlist + i;

		if ((offset + 6 + 1 + strlen(nbr->ssid) + 2) > reserve_len)
			goto error;

		memcpy(&t->data[offset], nbr->bssid, 6);	/* bssid */
		offset += 6;
		t->data[offset++] = strlen(nbr->ssid);		/* ssid len */
		/* ssid */
		memcpy(&t->data[offset], nbr->ssid, strlen(nbr->ssid));
		offset += strlen(nbr->ssid);
		t->data[offset++] = nbr->rssi;			/* rcpi */

		snprintf(bw_str, 15, "%d", nbr->bw);
		t->data[offset++] = strlen(bw_str);		/* BW length */
		if ((offset + strlen(bw_str) + 1) > reserve_len)
			goto error;

		memcpy(&t->data[offset], bw_str, strlen(bw_str));
		offset += strlen(bw_str);
		t->data[offset++] = bssload_elem_pres;		/* BSS load element */

		if (bssload_elem_pres & CH_SCAN_RESULT_BSSLOAD_PRESENT) {
			if ((offset + 1 + 2) > reserve_len)
				goto error;

			t->data[offset++] = nbr->utilization;			/* channel utilization */
			BUF_PUT_BE16(t->data[offset], nbr->num_stations);	/* station count */
			offset += 2;
		}
	}

	/* TODO/FIXME
	 * add the total scan duration for active acan
	 */
	if ((offset + 5) > reserve_len)
		goto error;

	BUF_PUT_BE32(t->data[offset], 0);		/* scan duration */
	offset += 4;
	t->data[offset++] = SCAN_RESULT_SCAN_TYPE;	/* scan type */

out:
	/* update the tlv length */
	t->len = offset;

	ret = cmdu_put_tlv(cmdu, t);
	if (ret) {
		fprintf(stderr, "%s: error: cmdu_put_tlv()\n", __func__);
		return -1;
	}

	return 0;

error:
	return -1;
}

int deagent_gen_client_assoc_event(struct deagent_private *priv,
		struct cmdu_buff *cmdu, uint8_t *sta_mac,
		const char *event, uint8_t *bssid)
{
	int ret;
	struct tlv *t;
	struct tlv_client_assoc_event *data;

	t = cmdu_reserve_tlv(cmdu, 30);
	if (!t)
		return -1;

	t->type = MAP_TLV_CLIENT_ASSOCIATION_EVENT;
	t->len = sizeof(*data);
	data = (struct tlv_client_assoc_event *)t->data;
	memcpy(data->macaddr, sta_mac, 6);
	memcpy(data->bssid, bssid, 6);
	if (!strcmp(event, "connected"))
		data->event = 1 << 7;
	else if (!strcmp(event, "disconnected"))
		data->event = 0;

	ret = cmdu_put_tlv(cmdu, t);
	if (ret) {
		fprintf(stderr, "%s: error: cmdu_put_tlv()\n", __func__);
		return -1;
	}

	return 0;
}

int deagent_gen_sta_mac(struct deagent_private *priv,
		struct cmdu_buff *cmdu, uint8_t *sta)
{
	int ret;
	struct tlv *t;
	struct tlv_sta_mac *data;

	t = cmdu_reserve_tlv(cmdu, 20);
	if (!t)
		return -1;

	t->type = MAP_TLV_STA_MAC_ADDRESS;
	t->len = sizeof(*data);
	data = (struct tlv_sta_mac *)t->data;
	memcpy(data->macaddr, sta, 6);

	ret = cmdu_put_tlv(cmdu, t);
	if (ret) {
		fprintf(stderr, "%s: error: cmdu_put_tlv()\n", __func__);
		return -1;
	}

	return 0;
}

int deagent_gen_reason_code(struct deagent_private *priv,
		struct cmdu_buff *cmdu, int reason_code)
{
	int ret;
	struct tlv *t;
	struct tlv_reason_code *data;

	t = cmdu_reserve_tlv(cmdu, 20);
	if (!t)
		return -1;

	t->type = MAP_TLV_REASON_CODE;
	t->len = sizeof(*data);
	data = (struct tlv_reason_code *)t->data;
	BUF_PUT_BE16(data->code, reason_code & 0xffff);

	ret = cmdu_put_tlv(cmdu, t);
	if (ret) {
		fprintf(stderr, "%s: error: cmdu_put_tlv()\n", __func__);
		return -1;
	}

	return 0;
}

int deagent_gen_disassoc_stats(struct deagent_private *priv, struct cmdu_buff *cmdu,
		uint8_t *sta, struct wifi_disassoc_events *disassoc_ev)
{
	int ret;
	struct tlv *t;
	struct tlv_assoc_sta_traffic_stats *data;

	if (!disassoc_ev)
		return -1;

	t = cmdu_reserve_tlv(cmdu, 64);
	if (!t)
		return -1;

	t->type = MAP_TLV_ASSOCIATED_STA_TRAFFIC_STATS;
	t->len = sizeof(*data);
	data = (struct tlv_assoc_sta_traffic_stats *)t->data;

	memcpy(data->macaddr, sta, 6);
	BUF_PUT_BE32(data->tx_bytes, disassoc_ev->sta_data.tx_bytes);
	BUF_PUT_BE32(data->rx_bytes, disassoc_ev->sta_data.rx_bytes);
	BUF_PUT_BE32(data->tx_packets, disassoc_ev->sta_data.tx_pkts);
	BUF_PUT_BE32(data->rx_packets, disassoc_ev->sta_data.rx_pkts);
	BUF_PUT_BE32(data->tx_err_packets, disassoc_ev->sta_data.tx_errors);
	BUF_PUT_BE32(data->tx_err_packets, disassoc_ev->sta_data.rx_errors);
	BUF_PUT_BE32(data->rtx_packets, disassoc_ev->sta_data.rtx_pkts);

	ret = cmdu_put_tlv(cmdu, t);
	if (ret) {
		fprintf(stderr, "%s: error: cmdu_put_tlv()\n", __func__);
		return -1;
	}

	return 0;
}

int deagent_gen_client_info(struct deagent_private *priv,
		struct cmdu_buff *cmdu, uint8_t *sta,
		uint8_t *bssid)
{
	int ret;
	struct tlv *t;
	struct tlv_client_info *data;

	t = cmdu_reserve_tlv(cmdu, 256);
	if (!t)
		return -1;

	t->type = MAP_TLV_CLIENT_INFO;
	t->len = sizeof(*data);
	data = (struct tlv_client_info *)t->data;
	memcpy(data->bssid, bssid, 6);
	memcpy(data->macaddr, sta, 6);

	ret = cmdu_put_tlv(cmdu, t);
	if (ret) {
		fprintf(stderr, "%s: error: cmdu_put_tlv()\n", __func__);
		return -1;
	}

	return 0;
}

int deagent_gen_client_cap_report(struct deagent_private *priv,
		struct cmdu_buff *cmdu, uint8_t result,
		struct wifi_assoc_frame *assoc_frame)
{
	int ret;
	struct tlv *t;
	struct tlv_client_cap_report *data;

	if (result != 0x00) {
		t = cmdu_reserve_tlv(cmdu, 10);
		if (!t)
			return -1;

		t->type = MAP_TLV_CLIENT_CAPABILITY_REPORT;
		t->len = sizeof(*data);
		data = (struct tlv_client_cap_report *)t->data;

		data->result = result;
	} else {
		/* [re]assoc frame received without the header
		 * so no need to skip bytes
		 */
		int len;
		uint8_t *body;

		len = assoc_frame->len;
		body = assoc_frame->frame;

		/* reseve addition few bytes (say 10 more bytes)
		 * actual required: len + 3; (1: tlv_type, 2: tlv_len)
		 */
		t = cmdu_reserve_tlv(cmdu, len + 10);
		if (!t)
			return -1;

		t->type = MAP_TLV_CLIENT_CAPABILITY_REPORT;
		t->len = sizeof(*data);
		data = (struct tlv_client_cap_report *)t->data;
		data->result = result;

		/* copy the raw frame */
		memcpy(data->frame, body, len);
		t->len += len;
	}

	ret = cmdu_put_tlv(cmdu, t);
	if (ret) {
		fprintf(stderr, "%s: error: cmdu_put_tlv()\n", __func__);
		return -1;
	}

	return 0;
}

int deagent_gen_error_code(struct deagent_private *priv, struct cmdu_buff *cmdu,
		uint8_t reason, uint8_t *sta)
{
	int ret;
	struct tlv *t;
	struct tlv_error_code *data;

	t = cmdu_reserve_tlv(cmdu, 256);
	if (!t)
		return -1;

	t->type = MAP_TLV_ERROR_CODE;
	t->len = sizeof(*data);
	data = (struct tlv_error_code *)t->data;
	data->reason = reason;
	memcpy(data->macaddr, sta, 6);

	ret = cmdu_put_tlv(cmdu, t);
	if (ret) {
		fprintf(stderr, "%s: error: cmdu_put_tlv()\n", __func__);
		return -1;
	}

	return 0;
}

int deagent_gen_vendor_specific(struct wifi_network_device *dev,
		struct cmdu_buff *cmdu)
{
	int ret, i, j, k;
	int offset = 0;
	uint16_t cnt = 0;
	struct tlv *t;
	struct tlv_vendor_metric_specific {
		uint8_t oui[3];
		uint16_t num_sta;
		struct assoc_sta_metrics {
			uint8_t macaddr[6];
			uint64_t tx_bytes;
			uint64_t rx_bytes;
			uint64_t tx_packets;
			uint64_t rx_packets;
		} __attribute__((packed)) sta[];
	} __attribute__((packed));
	struct tlv_vendor_metric_specific *data;

	t = cmdu_reserve_tlv(cmdu, 512);
	if (!t)
		return -1;

	t->type = TLV_TYPE_VENDOR_SPECIFIC;
	offset = sizeof(*data);
	data= (struct tlv_vendor_metric_specific *)t->data;

	data->oui[0] = 0x00;
	data->oui[1] = 0x22;
	data->oui[2] = 0x07;

	for (i = 0; i < dev->num_radios; i++) {
		struct wifi_radio_element *radio = dev->radiolist + i;

		for (j = 0; j < radio->num_bss; j++) {
			struct wifi_bss_element *bss = radio->bsslist + j;

			for (k = 0; k < bss->num_stations; k++) {
				struct wifi_sta_element *sta = bss->stalist + k;
				struct assoc_sta_metrics *sm =
					(struct assoc_sta_metrics *)&t->data[offset];

				memcpy(sm->macaddr, sta->macaddr, 6);
				BUF_PUT_BE64(sm->tx_bytes, sta->tx_bytes);
				BUF_PUT_BE64(sm->rx_bytes, sta->rx_bytes);
				BUF_PUT_BE64(sm->tx_packets, sta->tx_pkts);
				BUF_PUT_BE64(sm->rx_packets, sta->rx_pkts);

				offset += sizeof(*sm);
				cnt++;
			}
		}
	}

	BUF_PUT_BE16(data->num_sta, cnt);
	t->len = offset;

	ret = cmdu_put_tlv(cmdu, t);
	if (ret) {
		fprintf(stderr, "%s: error: cmdu_put_tlv()\n", __func__);
		return -1;
	}

	return 0;
}
