/*
 * decollector.c
 * WiFi Data Elements plugin
 *
 * Copyright (C) 2020 IOPSYS Software Solutions AB. All rights reserved.
 *
 * Author: anjan.chanda@iopsys.eu
 *
 * See LICENSE file for license related information.
 *
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <libubus.h>
#include <libubox/utils.h>

#include <easy/easy.h>
#include <1905_tlvs.h>

#include "timer.h"
#include <cmdu_ackq.h>
#include <map_module.h>

#include "wfa_delm.h"
#include "decollector.h"
#include "debug.h"

void free_radio_events(struct decollector_private *p)
{
	int i;
	struct wifi_radio_events *rv = &p->rv;

	for (i = 0; i < rv->num_radios; i++) {
		struct radio_event_data *r = rv->rdata + i;

		if (r->num_cdata > 0)
			free(r->cdata);
	}

	if (rv->num_radios > 0)
		free(rv->rdata);
}

void decollector_clear_framelist(struct decollector_private *priv)
{
	struct wifi_assoc_frame *p = NULL, *tmp;

	list_for_each_entry_safe(p, tmp,  &priv->framelist, list) {
		list_del(&p->list);
		free(p->frame);
		free(p);
	}
}

void refresh_timer_cb(atimer_t *t)
{
	struct decollector_private *p = container_of(t, struct decollector_private, t);

	fprintf(stderr, "decollector: %s: periodic = '%u' ms.\n",
		__func__, p->cfg.refresh_int);

	if (!p->getter_running)
		decollector_getter_run(p);

	if (p->cfg.refresh_int)
		timer_set(t, p->cfg.refresh_int);
}

static int decollector_ackq_timeout_cb(struct cmdu_ackq *q, struct cmdu_ackq_entry *msg)
{
	struct decollector_private *p = container_of(q, struct decollector_private, cmdu_ack_q);
	struct wifi_network_device *dev = NULL;

	/* reduce the pending count & update the datamodel
	 */
	p->pending--;
	dev = decollector_get_origin_dev(p, msg->origin);
	if (dev) {
		/* TODO/REVISIT:
		 * mark the node as invalid node if response is not
		 * received for TOPOLOGY_QUERY, else
		 * mark the node as valid node to show the available data
		 */
		if (msg->type == CMDU_TYPE_TOPOLOGY_RESPONSE) {
			dev->is_valid_node = false;
			fprintf(stderr, "marking the node as invalid node: "
				MACFMT "\n", MAC2STR(dev->macaddr));
		} else {
			dev->is_valid_node = true;
			fprintf(stderr, "partial data has been received from this node: "
				MACFMT "\n", MAC2STR(dev->macaddr));
		}

		if (p->pending <= 0)
			decollector_update_datamodel(p);
	}

	return CMDU_ACKQ_TMO_DELETE;
}

int decollector_exit(void *priv)
{
	struct ubus_object *obj;
	struct ubus_object_type *obj_type;
	struct ubus_method *obj_methods;
	struct wifi_sta_events *sv = NULL;
	struct decollector_private *p = (struct decollector_private *)priv;

	if (p) {
		ubus_unregister_event_handler(p->ctx, &p->topology_ev);
		ubus_unregister_event_handler(p->ctx, &p->ubus_obj_ev);
		map_unsubscribe(p->ctx, p->subscriber);

		decollector_free_dm(p->dm);
		decollector_free_dm(p->dmi);

		obj = &p->obj;
		obj_methods = (struct ubus_method *)obj->methods;
		if (obj_methods)
			free(obj_methods);

		obj_type = obj->type;
		if (obj_type)
			free(obj_type);

		ubus_free(p->ctx);

		sv = p->sv;
		if (sv) {
			if (sv->num_assoc_events > 0)
				free(sv->assoclist);

			if (sv->num_disassoc_events > 0)
				free(sv->disassoclist);

			free(sv);
		}

		cmdu_ackq_free(&p->cmdu_ack_q);
		free(p);
	}

	return 0;
}

int decollector_init(void **priv, void *cfg)
{
	struct decollector_private *p;
	int ret;

	*priv = NULL;
	p = calloc(1, sizeof(struct decollector_private));
	if (!p)
		return -1;

	cmdu_ackq_init(&p->cmdu_ack_q);
	p->cmdu_ack_q.timeout_cb = decollector_ackq_timeout_cb;

	/* sta assoc frame list */
	INIT_LIST_HEAD(&p->framelist);

	p->dm = NULL;
	p->dmi = NULL;

	p->sv = calloc(1, sizeof(struct wifi_sta_events));
	if (!p->sv)
		goto out_err;

	if (cfg) {
		struct decollector_config *c = (struct decollector_config *)cfg;

		p->cfg.uctx = c->uctx;
		p->cfg.refresh_int = c->refresh_int;
		p->cfg.is_collector = 1;
		p->cfg.is_sap_mode = c->is_sap_mode;
		p->cfg.profile = c->profile;

		p->ctx = c->uctx;
	}


	fprintf(stderr, "%s:  priv = %p\n", __func__, p);

	p->i1905 = decollector_lookup_object(p, IEEE1905_OBJECT);
	if (p->i1905 == OBJECT_INVALID) {
		err("Failed to get '%s' object!\n", IEEE1905_OBJECT);
		goto out_err;
	}

	p->i1905_map = decollector_lookup_object(p, MAP_PLUGIN_OBJECT);
	if (p->i1905_map == OBJECT_INVALID) {
		err("Failed to get '%s' object!\n", MAP_PLUGIN_OBJECT);
		goto out_err;
	}

	ret = decollector_get_ieee1905id(p);
	if (ret)
		goto out_err;

	ret = decollector_register_events(p);
	if (ret)
		goto out_err;

	ret = decollector_publish_object(p, WFADATAD_COLLECTOR_OBJECT);
	if (ret)
		goto out_err;


	/* subscribe for CMDU */
	decollector_subscribe_for_cmdus(p);

	/* send AP-autoconfig search message to get the controller id */
	ret = prepare_autoconfig_search(p, 0x00);	/* 2.4 GHZ */
	if (ret)
		fprintf(stderr, "error in sending ap-autoconfig message\n");

	/* start DE-collector */
	decollector_getter_run(p);

	timer_init(&p->t, refresh_timer_cb);
	if (p->cfg.refresh_int)
		timer_set(&p->t, p->cfg.refresh_int);

	*priv = p;
	return 0;

out_err:
	cmdu_ackq_free(&p->cmdu_ack_q);
	decollector_clear_framelist(p);
	free(p);
	return -1;
}
