/*
 * wfa_delm.h
 * WiFi Alliance Data Elements v1.1 definitions header.
 *
 * Copyright (C) 2019 IOPSYS Software Solutions AB. All rights reserved.
 *
 * Author: anjan.chanda@iopsys.eu
 *
 */

#ifndef WFA_DELM_H
#define WFA_DELM_H

#include <stdint.h>
#include <stdbool.h>


#ifdef __cplusplus
extern "C" {
#endif

#ifndef WIFI_H
enum wifi_band {
	BAND_2       = 1 << 0,        /**< 0x1 for 2.4Ghz band. */
	BAND_5       = 1 << 1,        /**< 0x2 for 5Ghz band. */
	BAND_DUAL    = 3,             /**< 0x3 for both 2.4 and 5Ghz bands */
	BAND_60      = 1 << 2,        /**< 0x4 for 60Ghz */
	BAND_6       = 1 << 3,        /**< 0x8 for 6Ghz */
	BAND_UNKNOWN = 1 << 4,        /**< 0x10 (and above) for unknown band */
};

/** enum wifi_bw - bandwidth */
enum wifi_bw {
	BW20,             /**< bandwidth = 20 Mhz */
	BW40,             /**< bandwidth = 40 Mhz */
	BW80,             /**< bandwidth = 80 Mhz */
	BW160,            /**< bandwidth = 160 Mhz */
	BW_AUTO,          /**< bandwidth = ? */
	BW8080 = 7,       /**< bandwidth = 80+80 Mhz */
	BW_UNKNOWN,       /**< Unknown or Undefined */
};

/** enum wifi_chan_ext - extension channel */
enum wifi_chan_ext {
	EXTCH_NONE,     /**< no extension channel */
	EXTCH_ABOVE,    /**< extension channel above primary */
	EXTCH_BELOW,    /**< extension channel below primary */
	EXTCH_AUTO,     /**< extension channel auto */
};

enum wifi_security_type {
	WIFI_SECURITY_NONE,
	WIFI_SECURITY_WEP64,
	WIFI_SECURITY_WEP128,
	WIFI_SECURITY_WPAPSK,
	WIFI_SECURITY_WPA2PSK,
	WIFI_SECURITY_FT_WPA2PSK,
	WIFI_SECURITY_WPA3PSK,
	WIFI_SECURITY_WPA3PSK_T,
	WIFI_SECURITY_FT_WPA3PSK,
	WIFI_SECURITY_WPA,
	WIFI_SECURITY_WPA2,
	WIFI_SECURITY_WPA3,
};
#endif

enum aptype {
	AP_WIFI_FBSS,
	AP_WIFI_BBSS,
	AP_WIFI_COMBINED,
};


typedef char timestamp_t[32];
typedef uint8_t guid_t[16];
typedef uint8_t macaddr_t[6], node_id_t[6];

typedef struct ipaddress {
	uint8_t ver;
	uint8_t addr[16];
} ipaddr_t;


struct wifi_mcs {
	uint16_t vht_mcs_rxmap;
	uint16_t vht_mcs_txmap;
};

struct wifi_caps_element {
#define HT_CAP_VALID       0x4
#define VHT_CAP_VALID      0x8
#define HE_CAP_VALID       0x20
	uint8_t ht;
	uint8_t vht[6];
	/* TODO/revisit;
	 * 1 byte: length for supported MCS field
	 * 12 byte: HE Tx Rx MCS
	 * 2 btye: other fields
	 */
	uint8_t he[15];
	uint32_t valid;		/* caps validity */
	struct wifi_mcs mcs;
};

struct wifi_opclass_supported_element {
	uint8_t id;                     /* class number */
	int8_t max_txpower;
	uint32_t num_exclude_channels;
	uint8_t *exclude_chanlist;      /* list of non-operable channels */
	uint32_t num_channels;
	uint8_t *chanlist;
};

struct wifi_opclass_current_element {
	timestamp_t tsp;
	uint8_t id;
	uint8_t channel;
	int8_t txpower;
};

struct wifi_sta_element {
	timestamp_t tsp;
	uint8_t macaddr[6];
	struct wifi_caps_element caps;
	uint32_t dl_rate;             /* latest data rate in Kbps: ap -> sta */
	uint32_t ul_rate;             /* latest data rate in Kbps: sta -> ap */
	unsigned long ul_utilization; /* time in msecs for receive from sta */
	unsigned long dl_utilization; /* time in msecs for transmit to sta */
	uint32_t dl_est_thput;        /* in Mbps */
	uint32_t ul_est_thput;        /* in Mbps */
	uint8_t rssi;                 /* in dBm */
	uint32_t conn_time;           /* in secs since last associated */
	uint64_t tx_bytes;            /* transmit bytes count: ap -> sta */
	uint64_t rx_bytes;            /* receive bytes count: sta -> ap */
	uint32_t tx_pkts;
	uint32_t rx_pkts;
	uint32_t tx_errors;
	uint32_t rx_errors;
	uint32_t rtx_pkts;            /* total retransmitted packets */
	ipaddr_t ipaddr;
	char hostname[128];
	uint8_t num_bcn_reports;
	uint8_t *bcn_reportlist;      /* list of beacon reports */
};

struct wifi_bss_element {
	timestamp_t tsp;
	uint8_t bssid[6];
	char ssid[33];
	bool enabled;
	uint32_t uptime;
	uint64_t tx_ucast_bytes;
	uint64_t rx_ucast_bytes;
	uint64_t tx_mcast_bytes;
	uint64_t rx_mcast_bytes;
	uint64_t tx_bcast_bytes;
	uint64_t rx_bcast_bytes;
	uint8_t is_ac_be;
	uint8_t is_ac_bk;
	uint8_t is_ac_vo;
	uint8_t is_ac_vi;
	uint8_t est_wmm_be[3];
	uint8_t est_wmm_bk[3];
	uint8_t est_wmm_vi[3];
	uint8_t est_wmm_vo[3];
	uint32_t num_stations;
	struct wifi_sta_element *stalist;
};

struct wifi_scanres_neighbor_element {
	uint8_t bssid[6];
	char ssid[33];
	int rssi;
	uint32_t bw;
	uint8_t utilization;
	uint32_t num_stations;
};

struct wifi_scanres_channel_element {
	timestamp_t tsp;
	uint8_t channel;
	uint8_t utilization;
	uint8_t anpi;
	uint32_t num_neighbors;
	struct wifi_scanres_neighbor_element *nbrlist;  /* scanned AP list */
};

struct wifi_scanres_opclass_element {
	uint8_t opclass;
	uint32_t num_channels_scanned;
	struct wifi_scanres_channel_element *channel_scanlist;
};

struct wifi_scanres_element {
	timestamp_t tsp;
	uint32_t num_opclass_scanned;
	struct wifi_scanres_opclass_element *opclass_scanlist;
};

struct wifi_backhaul_element {
	uint8_t macaddr[6];
};

struct wifi_unassoc_sta_element {
	uint8_t macaddr[6];
	uint8_t rssi;
};

struct wifi_radio_element {
	uint8_t macaddr[6];
	bool enabled;
	uint8_t anpi;
	uint8_t total_utilization;   /** in %age, linearly scaled 0..255 */
	uint8_t tx_utilization;
	uint8_t rx_utilization;
	uint8_t other_utilization;
	uint8_t tx_streams;
	uint8_t rx_streams;
	char country_code[4];

	uint32_t num_supp_opclass;
	uint32_t num_curr_opclass;
	uint32_t num_unassoc_sta;
	uint32_t num_bss;
	uint32_t num_scanresult;
	struct wifi_backhaul_element bksta;

	struct wifi_caps_element caps;
	struct wifi_opclass_supported_element *supp_opclass;
	struct wifi_opclass_current_element *curr_opclass;

	struct wifi_bss_element *bsslist;
	struct wifi_scanres_element *scanlist;
	struct wifi_unassoc_sta_element *unassoc_stalist;
};

struct wifi_network_device {
	bool is_valid_node;        /* node validity; true: valid, false: invalid */
	uint8_t macaddr[6];        /* unique device-id: eui-48 address */
	uint8_t map_caps[2];       /* multi-ap caps */
	uint32_t collect_int;      /* data collection interval in msecs */
	uint32_t map_profile;      /* MAP-Profile; 1: profile-1, 2: profile-2 */
	uint32_t num_radios;
	struct wifi_radio_element *radiolist;
	uint8_t *ieee1905_tlv_data;/* ieee1905 specific tlv data */
	int ieee1905_tlv_datalen;  /* ieee1905 tlv data len */
};

struct wifi_network_ssid {
	bool enabled;
	size_t ssidlen;
	uint8_t ssid[33];
	uint32_t band;
	uint32_t security;
	uint16_t vid;
	uint8_t multi_ap;
	enum aptype type;
	struct list_head list;
};

struct wifi_network {
	timestamp_t tsp;
	char id[16];               /* network id: guid */
	uint8_t cntlr_id[6];       /* controller id: macaddress */
	uint32_t num_ssid;
	struct list_head ssidlist; /* list of wifi_network_ssid */
	uint16_t primary_vid;
	uint16_t default_pcp;
	uint32_t num_devices;
	struct wifi_network_device device[];
};

struct wifi_data_element {
	timestamp_t tsp;
	char version[32];
	struct wifi_network network;
};

struct wifi_assoc_events {
	timestamp_t tsp;
	uint8_t macaddr[6];
	uint8_t bssid[6];
	uint16_t status_code;
	struct wifi_caps_element caps;
};

struct wifi_disassoc_events {
	timestamp_t tsp;
	uint8_t bssid[6];
	uint16_t reason_code;
	struct wifi_sta_element sta_data;
};

struct wifi_sta_events {
	uint32_t old_assoc_events;
	uint32_t old_disassoc_events;
	uint32_t num_assoc_events;
	uint32_t num_disassoc_events;
	struct wifi_assoc_events *assoclist;
	struct wifi_disassoc_events *disassoclist;
};

struct wifi_channel_data {
	uint8_t channel;
	uint8_t utilization;
	uint8_t anpi;
};

struct radio_event_data {
	uint8_t macaddr[6];
	uint32_t num_cdata;
	struct wifi_channel_data *cdata;
};

struct wifi_radio_events {
	uint32_t num_radios;
	struct radio_event_data *rdata;
};

#ifdef __cplusplus
}
#endif
#endif /* WFA_DELM_H */
